#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    Lion Krischer (lionkrischer@gmail.com), 2017
:license:
    MIT License
"""
import os
import platform

import numpy as np
import pytest


from pyexodus import exodus


_p = [
    {"io_size": 4, "word_size": 4, "f_dtype": np.float32},
    {"io_size": 8, "word_size": 8, "f_dtype": np.float64},
]

if platform.architecture()[0] == "64bit":  # pragma: no cover
    _p.append(
        {"io_size": 0, "word_size": 8, "f_dtype": np.float64},
    )
else:  # pragma: no cover
    _p.append(
        {"io_size": 0, "word_size": 4, "f_dtype": np.float32},
    )


@pytest.fixture(params=_p, ids=["io_size_%i" % _i["io_size"] for _i in _p])
def io_size(request):
    """
    Fixture to parametrize over the io_sizes.
    """
    return request.param


def test_get_side_set_names(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    # Put two side sets. Note that the ids of the side sets have little to
    # do with their naming inside the file...
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ) as e:
        e.put_side_set_params(4, 5, 0)
        e.put_side_set(
            4, np.ones(5, dtype=np.int32) * 2, np.ones(5, dtype=np.int32) * 3
        )
        e.put_side_set_name(4, "edge of the world")

        e.put_side_set_params(2, 5, 0)
        e.put_side_set(
            2, np.ones(5, dtype=np.int32) * 7, np.ones(5, dtype=np.int32) * 8
        )
        e.put_side_set_name(2, "hallo")

    with exodus(filename, mode="r") as e:
        assert e.get_side_set_names() == ["edge of the world", "hallo"]


def test_read_global_variables(tmpdir, io_size):

    filename = os.path.join(tmpdir.strpath, "example.e")
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ) as e:

        e.set_global_variable_number(1)
        e.put_global_variable_name("Test", 1)
        e.put_global_variable_value("Test", 1, 1.1)

    with exodus(filename, mode="r") as e:
        assert e.get_global_variable_names() == ["Test"]
        value = e.get_global_variable_values("Test")
        np.testing.assert_almost_equal(value, 1.1)


def test_get_side_set_ids(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    # Put two side sets. Note that the ids of the side sets have little to
    # do with their naming inside the file...
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ) as e:
        e.put_side_set_params(4, 5, 0)
        e.put_side_set(
            4, np.ones(5, dtype=np.int32) * 2, np.ones(5, dtype=np.int32) * 3
        )
        e.put_side_set_name(4, "edge of the world")

        e.put_side_set_params(2, 5, 0)
        e.put_side_set(
            2, np.ones(5, dtype=np.int32) * 7, np.ones(5, dtype=np.int32) * 8
        )
        e.put_side_set_name(2, "hallo")

    with exodus(filename, mode="r") as e:
        assert e.get_side_set_ids() == [4, 2]


def test_get_side_set_hex(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    # Hex elements.
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.arange(5, dtype=np.float64) * 3,
        )
        e.put_elem_blk_info(1, "HEX", 6, 8, 0)
        e.put_elem_connectivity(1, np.arange(6 * 8) + 7)
        e.put_side_set_params(4, 5, 0)
        e.put_side_set(
            4,
            np.arange(5, dtype=np.int32) + 2,
            np.arange(5, dtype=np.int32) + 1,
        )

    with exodus(filename, mode="r") as e:
        elem_idx, side_ids = e.get_side_set(id=4)

    np.testing.assert_equal(elem_idx, [2, 3, 4, 5, 6])
    np.testing.assert_equal(side_ids, [1, 2, 3, 4, 5])


def test_get_side_set_quad(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=2,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )
        e.put_elem_blk_info(1, b"QUAD", 6, 4, 0)
        e.put_elem_connectivity(1, np.arange(6 * 4) + 7)
        e.put_side_set_params(4, 4, 0)
        e.put_side_set(
            4,
            np.arange(4, dtype=np.int32) + 2,
            np.arange(4, dtype=np.int32) + 1,
        )

    with exodus(filename, mode="r") as e:
        elem_idx, side_ids = e.get_side_set(id=4)

    np.testing.assert_equal(elem_idx, [2, 3, 4, 5])
    np.testing.assert_equal(side_ids, [1, 2, 3, 4])


def test_get_side_set_invalid_id(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    # Put two side sets. Note that the ids of the side sets have little to
    # do with their naming inside the file...
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ) as e:
        e.put_side_set_params(4, 5, 0)
        e.put_side_set(
            4, np.ones(5, dtype=np.int32) * 2, np.ones(5, dtype=np.int32) * 3
        )
        e.put_side_set_name(4, "edge of the world")

        e.put_side_set_params(2, 5, 0)
        e.put_side_set(
            2, np.ones(5, dtype=np.int32) * 7, np.ones(5, dtype=np.int32) * 8
        )
        e.put_side_set_name(2, "hallo")

    with exodus(filename, mode="r") as e:
        with pytest.raises(ValueError) as f:
            e.get_side_set(id=7)

    assert (
        f.value.args[0]
        == "No side set with id 7 in file. Available ids: 4, 2."
    )


def test_get_side_set_node_list_hex(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    # Hex elements.
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.arange(5, dtype=np.float64) * 3,
        )
        e.put_elem_blk_info(1, "HEX", 6, 8, 0)
        e.put_elem_connectivity(1, np.arange(6 * 8) + 7)
        e.put_side_set_params(4, 5, 0)
        e.put_side_set(
            4,
            np.arange(5, dtype=np.int32) + 2,
            np.arange(5, dtype=np.int32) + 1,
        )

    with exodus(filename, mode="r") as e:
        num_nodes, local_node_ids = e.get_side_set_node_list(id=4)

    np.testing.assert_equal(num_nodes, [4, 4, 4, 4, 4])
    np.testing.assert_equal(
        local_node_ids,
        [
            15,
            16,
            20,
            19,
            24,
            25,
            29,
            28,
            33,
            34,
            38,
            37,
            39,
            43,
            46,
            42,
            47,
            50,
            49,
            48,
        ],
    )


def test_get_side_set_node_list_quad(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=2,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )
        e.put_elem_blk_info(1, b"QUAD", 6, 4, 0)
        e.put_elem_connectivity(1, np.arange(6 * 4) + 7)
        e.put_side_set_params(4, 4, 0)
        e.put_side_set(
            4,
            np.arange(4, dtype=np.int32) + 2,
            np.arange(4, dtype=np.int32) + 1,
        )

    with exodus(filename, mode="r") as e:
        num_nodes, local_node_ids = e.get_side_set_node_list(id=4)

    np.testing.assert_equal(num_nodes, [2, 2, 2, 2])
    np.testing.assert_equal(local_node_ids, [11, 12, 16, 17, 21, 22, 26, 23])


def test_get_coord_3d(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.arange(5, dtype=np.float64) * 3,
        )

    with exodus(filename, mode="r") as e:
        with pytest.raises(ValueError) as f:
            e.get_coord(0)
        assert f.value.args[0] == "Invalid index. Coordinate bounds: [1, 5]."
        with pytest.raises(ValueError) as f:
            e.get_coord(6)
        assert f.value.args[0] == "Invalid index. Coordinate bounds: [1, 5]."

        np.testing.assert_allclose(e.get_coord(1), [0.0, 0.0, 0.0])
        np.testing.assert_allclose(e.get_coord(2), [1.0, 2.0, 3.0])
        np.testing.assert_allclose(e.get_coord(3), [2.0, 4.0, 6.0])

        # Now all at once.
        np.testing.assert_allclose(
            e.get_coord([1, 2, 3]),
            [[0.0, 1.0, 2.0], [0.0, 2.0, 4.0], [0.0, 3.0, 6.0]],
        )


def test_get_coord_2d(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=2,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )

    with exodus(filename, mode="r") as e:
        with pytest.raises(ValueError) as f:
            e.get_coord(0)
        assert f.value.args[0] == "Invalid index. Coordinate bounds: [1, 5]."
        with pytest.raises(ValueError) as f:
            e.get_coord(6)
        assert f.value.args[0] == "Invalid index. Coordinate bounds: [1, 5]."

        np.testing.assert_allclose(e.get_coord(1), [0.0, 0.0, 0.0])
        np.testing.assert_allclose(e.get_coord(2), [1.0, 2.0, 0.0])
        np.testing.assert_allclose(e.get_coord(3), [2.0, 4.0, 0.0])


def test_get_coords_3d(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.arange(5, dtype=np.float64) * 3,
        )

    with exodus(filename, mode="r") as e:
        np.testing.assert_allclose(
            e.get_coords(), [np.arange(5), np.arange(5) * 2, np.arange(5) * 3]
        )


def test_get_coords_2d(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=2,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )

    with exodus(filename, mode="r") as e:
        np.testing.assert_allclose(
            e.get_coords(), [np.arange(5), np.arange(5) * 2, np.zeros(5)]
        )


def test_get_elem_connectivity_quad(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    # Generate test file.
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=2,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )
        e.put_elem_blk_info(1, "QUAD", 6, 4, 0)
        e.put_elem_connectivity(1, np.arange(6 * 4) + 7)

    with exodus(filename, mode="r") as e:
        conn, num_elem, num_nodes_per_elem = e.get_elem_connectivity(id=1)

    np.testing.assert_equal(conn, (np.arange(6 * 4) + 7).reshape((6, 4)))
    assert num_elem == 6
    assert num_nodes_per_elem == 4


def test_get_elem_connectivity_hex(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    # Generate test file.
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=2,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )
        e.put_elem_blk_info(1, "HEX", 3, 8, 0)
        e.put_elem_connectivity(1, np.arange(3 * 8) + 7)

    with exodus(filename, mode="r") as e:
        conn, num_elem, num_nodes_per_elem = e.get_elem_connectivity(id=1)

    np.testing.assert_equal(conn, (np.arange(3 * 8) + 7).reshape((3, 8)))
    assert num_elem == 3
    assert num_nodes_per_elem == 8


def test_get_elem_connectivity_only_some_indices(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    # Generate test file.
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=2,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )
        e.put_elem_blk_info(1, "HEX", 3, 8, 0)
        e.put_elem_connectivity(1, np.arange(3 * 8) + 7)

    with exodus(filename, mode="r") as e:
        conn, num_elem, num_nodes_per_elem = e.get_elem_connectivity(
            id=1, indices=[1, 3]
        )

    np.testing.assert_equal(
        conn, (np.arange(3 * 8) + 7).reshape((3, 8))[[0, 2]]
    )
    assert num_elem == 3
    assert num_nodes_per_elem == 8


def test_num_dims_accessor(tmpdir, io_size):
    # 2D
    filename = os.path.join(tmpdir.strpath, "example_2d.e")
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=2,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )
    with exodus(filename, mode="r") as e:
        assert e.num_dims == 2

    # 3D
    filename = os.path.join(tmpdir.strpath, "example_3d.e")
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5, dtype=np.float64),
            yCoords=np.arange(5, dtype=np.float64) * 2,
            zCoords=np.zeros(5),
        )
    with exodus(filename, mode="r") as e:
        assert e.num_dims == 3


def test_get_element_variable_values(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_element_variable_number(5)
    e.put_element_variable_name("random", 3)
    # requires an actual element block.
    e.put_elem_blk_info(1, "HEX", 6, 3, 0)
    e.put_element_variable_values(1, "random", 1, np.arange(6))
    e.close()

    with exodus(filename, mode="a") as e:
        values = e.get_element_variable_values(1, "random", 1)
        np.testing.assert_equal(values, np.arange(6))

        # Raises a value error if the variable does not exist.
        with pytest.raises(ValueError):
            e.get_element_variable_values(1, "rando", 1)


def test_get_node_variable_values(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_node_variable_number(2)
    e.put_node_variable_name("good friend", 1)
    e.put_node_variable_values("good friend", 1, np.arange(5))

    np.testing.assert_equal(
        actual=e.get_node_variable_values(name="good friend", step=1),
        desired=np.arange(5),
    )

    # Invalid step.
    with pytest.raises(ValueError) as err:
        e.get_node_variable_values(name="good friend", step=0)
    assert err.value.args[0] == "Step must be larger than zero."

    with pytest.raises(ValueError) as err:
        e.get_node_variable_values(name="good friend", step=10)
    assert err.value.args[0] == "Step must be 0 < step <= 1."

    # Invalid name.
    with pytest.raises(ValueError) as err:
        e.get_node_variable_values(name="random", step=1)
    assert err.value.args[0] == "'random' is not in list"


def test_get_node_variable_names(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_node_variable_number(2)
    e.put_node_variable_name("good friend", 1)

    assert e.get_node_variable_names() == ["good friend", ""]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    Lion Krischer (lionkrischer@gmail.com), 2022
:license:
    MIT License
"""
import numpy as np


def get_dim_size(f, dimension_name):
    """
    Compatibility for h5netcdf<0.14.0 and newer versions.

    :param f: The parameter.
    :param dimension_name: The name of the dimension.
    """
    dim = f.dimensions[dimension_name]
    # Used to be an integer.
    if isinstance(dim, (int, np.integer)):
        return dim
    elif dim is None:
        return dim
    # An object in newer version.
    elif dim.isunlimited():
        return None
    else:
        return dim.size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    Lion Krischer (lionkrischer@gmail.com), 2016
:license:
    MIT License
"""
import os
import platform

import h5netcdf
import numpy as np
import pytest

from pyexodus import exodus
from pyexodus.compat import get_dim_size

_p = [
    {"io_size": 4, "word_size": 4, "f_dtype": np.float32},
    {"io_size": 8, "word_size": 8, "f_dtype": np.float64},
]

if platform.architecture()[0] == "64bit":  # pragma: no cover
    _p.append(
        {"io_size": 0, "word_size": 8, "f_dtype": np.float64},
    )
else:  # pragma: no cover
    _p.append(
        {"io_size": 0, "word_size": 4, "f_dtype": np.float32},
    )


@pytest.fixture(params=_p, ids=["io_size_%i" % _i["io_size"] for _i in _p])
def io_size(request):
    """
    Fixture to parametrize over the io_sizes.
    """
    return request.param


def to_string(x):
    return x.decode() if isinstance(x, bytes) else x


def normalize(v):
    v = to_string(v)
    if isinstance(v, np.ndarray) and len(v) == 1:
        v = v[0]
    if hasattr(h5netcdf, "Dimension") and isinstance(v, h5netcdf.Dimension):
        if v.isunlimited() is False:
            v = v.size
        elif v.size == 0:
            v = None
        else:
            breakpoint()
            pass
    return v


def norm_dict(d):
    """
    Attempt to get dicts the same, independent of the h5netcdf version.
    """
    return {normalize(k): normalize(v) for k, v in dict(d).items()}


def test_initialization(tmpdir, io_size):
    """
    Tests initialization.

    Test data has been generated by using the official exodus Python API.
    """
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.close()

    # Just manually test everything.
    with h5netcdf.File(filename, mode="r") as f:
        assert norm_dict(f.attrs) == norm_dict(
            {
                "api_version": np.array([7.05], dtype=np.float32),
                "file_size": np.array([1], dtype=np.int32),
                "floating_point_word_size": np.array(
                    [io_size["word_size"]], dtype=np.int32
                ),
                "int64_status": np.array([0], dtype=np.int32),
                "maximum_name_length": np.array([32], dtype=np.int32),
                "title": b"Example",
                "version": np.array([7.05], dtype=np.float32),
            }
        )

        assert norm_dict(dict(f.dimensions)) == {
            "four": 4,
            "len_line": 81,
            "len_name": 256,
            "len_string": 33,
            "num_dim": 3,
            "num_el_blk": 1,
            "num_elem": 6,
            "num_nodes": 5,
            "num_side_sets": 1,
            "time_step": None,
        }

        assert list(f.groups) == []

        # Testing the variables is a bit more effort.

        # Generate with
        # v = f.variables
        # {k: {"dimensions": v[k].dimensions,
        #      "shape": v[k].shape,
        #      "dtype": v[k].dtype,
        #      "attrs": dict(v[k].attrs),
        #      "data": v[k][:]} for k in f.variables.keys()} == \

        expected = {
            "coor_names": {
                "attrs": {},
                "data": np.array(
                    [[""] * 256, [""] * 256, [""] * 256], dtype="|S1"
                ),
                "dimensions": ("num_dim", "len_name"),
                "dtype": np.dtype("S1"),
                "shape": (3, 256),
            },
            "coordx": {
                "attrs": {},
                "data": np.array([0.0, 0.0, 0.0, 0.0, 0.0]),
                "dimensions": ("num_nodes",),
                "dtype": io_size["f_dtype"],
                "shape": (5,),
            },
            "coordy": {
                "attrs": {},
                "data": np.array([0.0, 0.0, 0.0, 0.0, 0.0]),
                "dimensions": ("num_nodes",),
                "dtype": io_size["f_dtype"],
                "shape": (5,),
            },
            "coordz": {
                "attrs": {},
                "data": np.array([0.0, 0.0, 0.0, 0.0, 0.0]),
                "dimensions": ("num_nodes",),
                "dtype": io_size["f_dtype"],
                "shape": (5,),
            },
            "eb_names": {
                "attrs": {},
                "data": np.array([[""] * 256], dtype="|S1"),
                "dimensions": ("num_el_blk", "len_name"),
                "dtype": np.dtype("S1"),
                "shape": (1, 256),
            },
            "eb_prop1": {
                "attrs": {"name": b"ID"},
                "data": np.array([-1], dtype=np.int32),
                "dimensions": ("num_el_blk",),
                "dtype": np.dtype("int32"),
                "shape": (1,),
            },
            "eb_status": {
                "attrs": {},
                "data": np.array([0], dtype=np.int32),
                "dimensions": ("num_el_blk",),
                "dtype": np.dtype("int32"),
                "shape": (1,),
            },
            "ss_names": {
                "attrs": {},
                "data": np.array([[""] * 256], dtype="|S1"),
                "dimensions": ("num_side_sets", "len_name"),
                "dtype": np.dtype("S1"),
                "shape": (1, 256),
            },
            "ss_prop1": {
                "attrs": {"name": b"ID"},
                "data": np.array([-1], dtype=np.int32),
                "dimensions": ("num_side_sets",),
                "dtype": np.dtype("int32"),
                "shape": (1,),
            },
            "ss_status": {
                "attrs": {},
                "data": np.array([0], dtype=np.int32),
                "dimensions": ("num_side_sets",),
                "dtype": np.dtype("int32"),
                "shape": (1,),
            },
            "time_whole": {
                "attrs": {},
                "data": np.array([], dtype=io_size["f_dtype"]),
                "dimensions": ("time_step",),
                "dtype": io_size["f_dtype"],
                "shape": (0,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_coords(tmpdir, io_size):
    """
    Tests the put_coords() method.
    """
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_info_records(info=[])

    # Use different dtypes on purpose to test the type conversions.
    e.put_coords(
        xCoords=np.arange(5, dtype=np.float32),
        yCoords=np.arange(5, dtype=np.int32) * 2,
        zCoords=np.arange(5, dtype=np.int64) * 3,
    )

    e.close()

    expected = {
        "coordx": {
            "attrs": {},
            "data": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
            "dimensions": ("num_nodes",),
            "dtype": io_size["f_dtype"],
            "shape": (5,),
        },
        "coordy": {
            "attrs": {},
            "data": np.array([0.0, 2.0, 4.0, 6.0, 8.0]),
            "dimensions": ("num_nodes",),
            "dtype": io_size["f_dtype"],
            "shape": (5,),
        },
        "coordz": {
            "attrs": {},
            "data": np.array([0.0, 3.0, 6.0, 9.0, 12.0]),
            "dimensions": ("num_nodes",),
            "dtype": io_size["f_dtype"],
            "shape": (5,),
        },
    }

    with h5netcdf.File(filename, mode="r") as f:
        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_elem_blk_info(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_info_records(info=[])

    # Use different dtypes on purpose to test the type conversions.
    e.put_coords(
        xCoords=np.arange(5, dtype=np.float32),
        yCoords=np.arange(5, dtype=np.int32) * 2,
        zCoords=np.arange(5, dtype=np.int64) * 3,
    )

    e.put_elem_blk_info(1, "HEX", 6, 3, 0)

    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        # Two new dimensions.
        assert get_dim_size(f, "num_el_in_blk1") == 6
        assert get_dim_size(f, "num_nod_per_el1") == 3

        # One new variable.
        expected = {
            "connect1": {
                "attrs": {"elem_type": b"HEX"},
                "data": np.zeros((6, 3), dtype=np.int32),
                "dimensions": ("num_el_in_blk1", "num_nod_per_el1"),
                "dtype": np.int32,
                "shape": (6, 3),
            },
            "eb_status": {
                "attrs": {},
                "data": np.ones(1),
                "dimensions": ("num_el_blk",),
                "dtype": np.int32,
                "shape": (1,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_elem_connectivity(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_info_records(info=[])

    # Use different dtypes on purpose to test the type conversions.
    e.put_coords(
        xCoords=np.arange(5, dtype=np.float32),
        yCoords=np.arange(5, dtype=np.int32) * 2,
        zCoords=np.arange(5, dtype=np.int64) * 3,
    )

    e.put_elem_blk_info(1, "HEX", 6, 3, 0)
    e.put_elem_connectivity(1, np.arange(6 * 3) + 7)

    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        # connect1 should now be filled.
        expected = {
            "connect1": {
                "attrs": {"elem_type": b"HEX"},
                "data": (np.arange(6 * 3) + 7).reshape((6, 3)),
                "dimensions": ("num_el_in_blk1", "num_nod_per_el1"),
                "dtype": np.int32,
                "shape": (6, 3),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_elem_connectivity_indices_shift(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_info_records(info=[])

    # Use different dtypes on purpose to test the type conversions.
    e.put_coords(
        xCoords=np.arange(5, dtype=np.float32),
        yCoords=np.arange(5, dtype=np.int32) * 2,
        zCoords=np.arange(5, dtype=np.int64) * 3,
    )

    e.put_elem_blk_info(1, "HEX", 6, 3, 0)
    e.put_elem_connectivity(1, np.arange(6 * 3) + 7, shift_indices=3)

    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        # connect1 should now be filled.
        expected = {
            "connect1": {
                "attrs": {"elem_type": b"HEX"},
                "data": (np.arange(6 * 3) + 7).reshape((6, 3)) + 3,
                "dimensions": ("num_el_in_blk1", "num_nod_per_el1"),
                "dtype": np.int32,
                "shape": (6, 3),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_time(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_time(1, 1.1)

    with h5netcdf.File(filename, mode="r") as f:
        np.testing.assert_allclose(f.variables["time_whole"], [1.1])


def test_set_global_variable_number(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_global_variable_number(3)

    with h5netcdf.File(filename, mode="r") as f:
        _d = np.empty((3, 256), dtype="|S1")
        _d.fill("")

        expected = {
            "name_glo_var": {
                "attrs": {},
                "data": _d,
                "dimensions": ("num_glo_var", "len_name"),
                "dtype": np.dtype("|S1"),
                "shape": (3, 256),
            },
            "vals_glo_var": {
                "attrs": {},
                "data": np.zeros((0, 3)),
                "dimensions": ("time_step", "num_glo_var"),
                "dtype": io_size["f_dtype"],
                "shape": (0, 3),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_global_variable_name(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_global_variable_number(3)
    e.put_global_variable_name(name="hello", index=2)

    _d = np.empty((3, 256), dtype="|S1")
    _d.fill("")
    _d[1][:5] = list("hello")

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "name_glo_var": {
                "attrs": {},
                "data": _d,
                "dimensions": ("num_glo_var", "len_name"),
                "dtype": np.dtype("|S1"),
                "shape": (3, 256),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_get_global_variable_names(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")
    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_global_variable_number(3)
    e.put_global_variable_name(name="hello", index=2)
    assert e.get_global_variable_names() == ["", "hello", ""]


def test_put_global_variable_value(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_global_variable_number(3)
    e.put_global_variable_name(name="hello", index=2)

    e.put_global_variable_value("hello", 1, 1.1)

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "vals_glo_var": {
                "attrs": {},
                "data": np.array([[0, 1.1, 0]], dtype=io_size["f_dtype"]),
                "dimensions": ("time_step", "num_glo_var"),
                "dtype": io_size["f_dtype"],
                "shape": (1, 3),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_set_element_variables(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    # This call does nothing.
    e.set_element_variable_number(0)

    # This does something.
    e.set_element_variable_number(5)
    e.close()

    _d = np.empty((5, 256), dtype="|S1")
    _d.fill("")

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "name_elem_var": {
                "attrs": {},
                "data": _d,
                "dimensions": ("num_elem_var", "len_name"),
                "dtype": np.dtype("|S1"),
                "shape": (5, 256),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_element_variable_name(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_element_variable_number(5)
    e.put_element_variable_name("random", 3)
    e.close()

    _d = np.empty((5, 256), dtype="|S1")
    _d.fill("")
    _d[2][:6] = list("random")

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "name_elem_var": {
                "attrs": {},
                "data": _d,
                "dimensions": ("num_elem_var", "len_name"),
                "dtype": np.dtype("|S1"),
                "shape": (5, 256),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_get_element_variable_names(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_element_variable_number(5)
    e.put_element_variable_name("random", 3)

    assert e.get_element_variable_names() == ["", "", "random", "", ""]


def test_put_element_variable_values(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_element_variable_number(5)
    e.put_element_variable_name("random", 3)
    # requires an actual element block.
    e.put_elem_blk_info(1, "HEX", 6, 3, 0)
    e.put_element_variable_values(1, "random", 1, np.arange(6))

    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "vals_elem_var3eb1": {
                "attrs": {},
                "data": np.arange(6, dtype=io_size["f_dtype"]).reshape(1, 6),
                "dimensions": ("time_step", "num_el_in_blk1"),
                "dtype": io_size["f_dtype"],
                "shape": (1, 6),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_set_node_variable_number(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_node_variable_number(2)

    _d = np.empty((2, 256), dtype="|S1")
    _d.fill("")

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "name_nod_var": {
                "attrs": {},
                "data": _d,
                "dimensions": ("num_nod_var", "len_name"),
                "dtype": np.dtype("|S1"),
                "shape": (2, 256),
            },
            "vals_nod_var1": {
                "attrs": {},
                "data": np.zeros((0, 5)),
                "dimensions": ("time_step", "num_nodes"),
                "dtype": io_size["f_dtype"],
                "shape": (0, 5),
            },
            "vals_nod_var2": {
                "attrs": {},
                "data": np.zeros((0, 5)),
                "dimensions": ("time_step", "num_nodes"),
                "dtype": io_size["f_dtype"],
                "shape": (0, 5),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_get_node_variable_number(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    assert e.get_node_variable_number() == 0
    e.set_node_variable_number(2)
    assert e.get_node_variable_number() == 2
    e.close()

    e = exodus(filename, mode="r")
    assert e.get_node_variable_number() == 2
    e.close()


def test_put_node_variable_name(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_node_variable_number(2)
    e.put_node_variable_name("good friend", 1)

    _d = np.empty((2, 256), dtype="|S1")
    _d.fill("")
    _d[0][:11] = list("good friend")

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "name_nod_var": {
                "attrs": {},
                "data": _d,
                "dimensions": ("num_nod_var", "len_name"),
                "dtype": np.dtype("|S1"),
                "shape": (2, 256),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_multiple_node_variable_names(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_node_variable_number(2)
    e.put_node_variable_name("good friend", 1)
    e.put_node_variable_name("how are you?", 2)

    _d = np.empty((2, 256), dtype="|S1")
    _d.fill("")
    _d[0][:11] = list("good friend")
    _d[1][:12] = list("how are you?")

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "name_nod_var": {
                "attrs": {},
                "data": _d,
                "dimensions": ("num_nod_var", "len_name"),
                "dtype": np.dtype("|S1"),
                "shape": (2, 256),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_node_variable_values(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.set_node_variable_number(2)
    e.put_node_variable_name("good friend", 1)
    e.put_node_variable_values("good friend", 1, np.arange(5))

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "vals_nod_var1": {
                "attrs": {},
                "data": np.arange(5).reshape((1, 5)),
                "dimensions": ("time_step", "num_nodes"),
                "dtype": io_size["f_dtype"],
                "shape": (1, 5),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_side_set_params(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_side_set_params(4, 5, 0)
    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "elem_ss1": {
                "attrs": {},
                "data": np.zeros(5),
                "dimensions": ("num_side_ss1",),
                "dtype": np.int32,
                "shape": (5,),
            },
            "side_ss1": {
                "attrs": {},
                "data": np.zeros(5),
                "dimensions": ("num_side_ss1",),
                "dtype": np.int32,
                "shape": (5,),
            },
            "ss_prop1": {
                "attrs": {"name": b"ID"},
                "data": np.array([4], dtype=np.int32),
                "dimensions": ("num_side_sets",),
                "dtype": np.dtype("int32"),
                "shape": (1,),
            },
            "ss_status": {
                "attrs": {},
                "data": np.array([1], dtype=np.int32),
                "dimensions": ("num_side_sets",),
                "dtype": np.dtype("int32"),
                "shape": (1,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_side_set(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_side_set_params(4, 5, 0)
    e.put_side_set(
        4, np.ones(5, dtype=np.int32) * 2, np.ones(5, dtype=np.int32) * 3
    )
    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "elem_ss1": {
                "attrs": {},
                "data": [2, 2, 2, 2, 2],
                "dimensions": ("num_side_ss1",),
                "dtype": np.int32,
                "shape": (5,),
            },
            "side_ss1": {
                "attrs": {},
                "data": [3, 3, 3, 3, 3],
                "dimensions": ("num_side_ss1",),
                "dtype": np.int32,
                "shape": (5,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_side_set_name(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_side_set_params(4, 5, 0)
    e.put_side_set(
        4, np.ones(5, dtype=np.int32) * 2, np.ones(5, dtype=np.int32) * 3
    )
    e.put_side_set_name(4, "edge of the world")
    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "ss_names": {
                "attrs": {},
                "data": np.array(
                    [
                        [
                            "e",
                            "d",
                            "g",
                            "e",
                            " ",
                            "o",
                            "f",
                            " ",
                            "t",
                            "h",
                            "e",
                            " ",
                            "w",
                            "o",
                            "r",
                            "l",
                            "d",
                        ]
                        + [""] * 239
                    ],
                    dtype="|S1",
                ),
                "dimensions": ("num_side_sets", "len_name"),
                "dtype": np.dtype("S1"),
                "shape": (1, 256),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_multiple_side_set_names(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    )
    e.put_side_set_params(4, 5, 0)
    e.put_side_set_params(7, 2, 0)
    e.put_side_set_name(4, "edge of the world")
    e.put_side_set_name(7, "a new beginning")
    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "ss_names": {
                "attrs": {},
                "data": np.array(
                    [
                        [
                            "e",
                            "d",
                            "g",
                            "e",
                            " ",
                            "o",
                            "f",
                            " ",
                            "t",
                            "h",
                            "e",
                            " ",
                            "w",
                            "o",
                            "r",
                            "l",
                            "d",
                        ]
                        + [""] * 239,
                        [
                            "a",
                            " ",
                            "n",
                            "e",
                            "w",
                            " ",
                            "b",
                            "e",
                            "g",
                            "i",
                            "n",
                            "n",
                            "i",
                            "n",
                            "g",
                        ]
                        + [""] * 241,
                    ],
                    dtype="|S1",
                ),
                "dimensions": ("num_side_sets", "len_name"),
                "dtype": np.dtype("S1"),
                "shape": (2, 256),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_usage_as_context_manager(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_time(1, 1.1)

    with h5netcdf.File(filename, mode="r") as f:
        np.testing.assert_allclose(f.variables["time_whole"], [1.1])


def test_status_in_file_with_two_side_sets_only_one_set(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ) as e:
        e.put_side_set_params(4, 5, 0)

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "ss_prop1": {
                "attrs": {"name": b"ID"},
                "data": np.array([4, -1], dtype=np.int32),
                "dimensions": ("num_side_sets",),
                "dtype": np.dtype("int32"),
                "shape": (2,),
            },
            "ss_status": {
                "attrs": {},
                "data": np.array([1, 0], dtype=np.int32),
                "dimensions": ("num_side_sets",),
                "dtype": np.dtype("int32"),
                "shape": (2,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_status_in_file_with_two_side_sets_with_two_set(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ) as e:
        e.put_side_set_params(4, 5, 0)
        e.put_side_set_params(7, 2, 0)

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "ss_prop1": {
                "attrs": {"name": b"ID"},
                "data": np.array([4, 7], dtype=np.int32),
                "dimensions": ("num_side_sets",),
                "dtype": np.dtype("int32"),
                "shape": (2,),
            },
            "ss_status": {
                "attrs": {},
                "data": np.array([1, 1], dtype=np.int32),
                "dimensions": ("num_side_sets",),
                "dtype": np.dtype("int32"),
                "shape": (2,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_filling_two_side_sets(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ) as e:
        e.put_side_set_params(4, 5, 0)
        e.put_side_set_params(7, 2, 0)
        e.put_side_set(
            4, np.ones(5, dtype=np.int32) * 2, np.ones(5, dtype=np.int32) * 3
        )
        e.put_side_set(
            7, np.ones(2, dtype=np.int32) * 7, np.ones(2, dtype=np.int32) * 8
        )

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "elem_ss1": {
                "attrs": {},
                "data": [2, 2, 2, 2, 2],
                "dimensions": ("num_side_ss1",),
                "dtype": np.int32,
                "shape": (5,),
            },
            "side_ss1": {
                "attrs": {},
                "data": [3, 3, 3, 3, 3],
                "dimensions": ("num_side_ss1",),
                "dtype": np.int32,
                "shape": (5,),
            },
            "elem_ss2": {
                "attrs": {},
                "data": [7, 7],
                "dimensions": ("num_side_ss2",),
                "dtype": np.int32,
                "shape": (2,),
            },
            "side_ss2": {
                "attrs": {},
                "data": [8, 8],
                "dimensions": ("num_side_ss2",),
                "dtype": np.int32,
                "shape": (2,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_put_info_records(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ) as e:
        e.put_info_records(["", "how", "are", "you"])

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "info_records": {
                "attrs": {},
                "data": np.array(
                    [
                        [
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                        ],
                        [
                            "h",
                            "o",
                            "w",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                        ],
                        [
                            "a",
                            "r",
                            "e",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                        ],
                        [
                            "y",
                            "o",
                            "u",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                            "",
                        ],
                    ],
                    dtype="|S1",
                ),
                "dimensions": ("num_info", "len_line"),
                "dtype": np.dtype("S1"),
                "shape": (4, 81),
            }
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert dict(a.attrs) == e["attrs"], key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_compression(tmpdir, io_size):
    """
    Test to make sure compressions does something.
    """
    filename = os.path.join(tmpdir.strpath, "example.e")

    # No compression at first.
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        io_size=io_size["io_size"],
    ):
        pass

    with h5netcdf.File(filename) as f:
        for var in f._variables:
            ds = f._variables[var]._h5ds
            assert ds.compression is None, ds
            assert ds.compression_opts is None, ds
            # Time arrays are resizeable so they must be chunked.
            if ds.name == "/time_whole":
                assert ds.chunks, ds
            else:
                assert ds.chunks is None, ds

    os.remove(filename)

    # No turn on compression.
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=2,
        compression=("gzip", 2),
        io_size=io_size["io_size"],
    ):
        pass

    with h5netcdf.File(filename) as f:
        for var in f._variables:
            ds = f._variables[var]._h5ds
            assert ds.compression == "gzip", ds
            assert ds.compression_opts == 2, ds
            # Compression forces chunking.
            assert ds.chunks, ds


def test_init_multiple_element_blocks(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=3,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "eb_names": {
                "attrs": {},
                "data": np.array(
                    [[""] * 256, [""] * 256, [""] * 256], dtype="|S1"
                ),
                "dimensions": ("num_el_blk", "len_name"),
                "dtype": np.dtype("S1"),
                "shape": (3, 256),
            },
            "eb_prop1": {
                "attrs": {"name": b"ID"},
                "data": np.array([-1, -1, -1], dtype=np.int32),
                "dimensions": ("num_el_blk",),
                "dtype": np.dtype("int32"),
                "shape": (3,),
            },
            "eb_status": {
                "attrs": {},
                "data": np.array([0, 0, 0], dtype=np.int32),
                "dimensions": ("num_el_blk",),
                "dtype": np.dtype("int32"),
                "shape": (3,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_init_multiple_element_blocks_and_set_first_one(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=3,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_elem_blk_info(7, "SEVEN", 6, 3, 0)
    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "eb_names": {
                "attrs": {},
                "data": np.array(
                    [[""] * 256, [""] * 256, [""] * 256], dtype="|S1"
                ),
                "dimensions": ("num_el_blk", "len_name"),
                "dtype": np.dtype("S1"),
                "shape": (3, 256),
            },
            "connect1": {
                "attrs": {"elem_type": b"SEVEN"},
                "data": np.zeros((6, 3), dtype=np.int32),
                "dimensions": ("num_el_in_blk1", "num_nod_per_el1"),
                "dtype": np.int32,
                "shape": (6, 3),
            },
            "eb_prop1": {
                "attrs": {"name": b"ID"},
                "data": np.array([7, -1, -1], dtype=np.int32),
                "dimensions": ("num_el_blk",),
                "dtype": np.dtype("int32"),
                "shape": (3,),
            },
            "eb_status": {
                "attrs": {},
                "data": np.array([1, 0, 0], dtype=np.int32),
                "dimensions": ("num_el_blk",),
                "dtype": np.dtype("int32"),
                "shape": (3,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_init_multiple_element_blocks_and_set_all(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")

    e = exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=3,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    )
    e.put_elem_blk_info(7, "SEVEN", 6, 3, 0)
    e.put_elem_blk_info(1, "ONE", 6, 3, 0)
    e.put_elem_blk_info(3, "THREE", 6, 3, 0)
    e.close()

    with h5netcdf.File(filename, mode="r") as f:
        expected = {
            "eb_names": {
                "attrs": {},
                "data": np.array(
                    [[""] * 256, [""] * 256, [""] * 256], dtype="|S1"
                ),
                "dimensions": ("num_el_blk", "len_name"),
                "dtype": np.dtype("S1"),
                "shape": (3, 256),
            },
            "connect1": {
                "attrs": {"elem_type": b"SEVEN"},
                "data": np.zeros((6, 3), dtype=np.int32),
                "dimensions": ("num_el_in_blk1", "num_nod_per_el1"),
                "dtype": np.int32,
                "shape": (6, 3),
            },
            "connect2": {
                "attrs": {"elem_type": b"ONE"},
                "data": np.zeros((6, 3), dtype=np.int32),
                "dimensions": ("num_el_in_blk2", "num_nod_per_el2"),
                "dtype": np.int32,
                "shape": (6, 3),
            },
            "connect3": {
                "attrs": {"elem_type": b"THREE"},
                "data": np.zeros((6, 3), dtype=np.int32),
                "dimensions": ("num_el_in_blk3", "num_nod_per_el3"),
                "dtype": np.int32,
                "shape": (6, 3),
            },
            "eb_prop1": {
                "attrs": {"name": b"ID"},
                "data": np.array([7, 1, 3], dtype=np.int32),
                "dimensions": ("num_el_blk",),
                "dtype": np.dtype("int32"),
                "shape": (3,),
            },
            "eb_status": {
                "attrs": {},
                "data": np.array([1, 1, 1], dtype=np.int32),
                "dimensions": ("num_el_blk",),
                "dtype": np.dtype("int32"),
                "shape": (3,),
            },
        }

        for key in sorted(expected.keys()):
            a = f.variables[key]
            e = expected[key]

            assert norm_dict(dict(a.attrs)) == norm_dict(e["attrs"]), key
            np.testing.assert_equal(a[:], e["data"], err_msg=key)
            assert a.dimensions == e["dimensions"], key
            assert a.dtype == e["dtype"], key
            assert a.shape == e["shape"], key


def test_get_elem_type_for_block(tmpdir, io_size):
    filename = os.path.join(tmpdir.strpath, "example.e")
    with exodus(
        filename,
        mode="w",
        title="Example",
        array_type="numpy",
        numDims=3,
        numNodes=5,
        numElems=6,
        numBlocks=1,
        numNodeSets=0,
        numSideSets=1,
        io_size=io_size["io_size"],
    ) as e:
        e.put_coords(
            xCoords=np.arange(5), yCoords=np.arange(5), zCoords=np.arange(5)
        )
        e.put_elem_blk_info(1, "HEX", 6, 3, 0)

        assert e.get_elem_type_for_block(1) == "HEX"

        with pytest.raises(ValueError) as err:
            e.get_elem_type_for_block(2)
        assert err.value.args[0] == "No element block with id 2 in file."



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    Lion Krischer (lionkrischer@gmail.com), 2022
:license:
    MIT License
"""
from __future__ import absolute_import

import os
import platform
import warnings

import numpy as np

import h5netcdf

from .compat import get_dim_size


# This uses zero based indexing to be compatible with numpy. The variables
# in the exodus files themselves are one based so keep that in mind!
# The values are from the exodus manual.
_SIDE_SET_NUMBERING = {
    "QUAD": np.array([[0, 1], [1, 2], [2, 3], [3, 0]], dtype=np.int32),
    "HEX": np.array(
        [
            [0, 1, 5, 4],
            [1, 2, 6, 5],
            [2, 3, 7, 6],
            [0, 4, 7, 3],
            [0, 3, 2, 1],
            [4, 5, 6, 7],
        ],
        dtype=np.int32,
    ),
}


class exodus(object):
    """
    Create a new Exodus file. Can also be used as a context manager.

    :type file: str
    :param file: Filename
    :type mode: str
    :param mode: File mode. Must currently be ``"r"``, ``"a"``, or ``"w"``.
    :type array_type: str
    :param array_type: Must be ``"numpy"``.
    :type title: str
    :param title: The title of the mesh.
    :type numDims: int
    :param numDims: The number of dimensions.
    :type numNodes: int
    :param numNodes: The number of nodes.
    :type numElems: int
    :param numElems: The number of elements.
    :type numBlocks: int
    :param numBlocks: The number of element blocks.
    :type numNodeSets: int
    :param numNodeSets: The number of node side sets.
    :type numSideSets: int
    :param numSideSets: The number of element side sets.
    :type io_size: int
    :param io_size: Determines how floating point variables are stored in
        the file.  Inputs will be converted if required.

        * ``0``: machine precision
        * ``4``: single precision
        * ``8``: double precision
    :type compression: tuple
    :param compression: Turn on compression. Pass a tuple of
        ``(method, option)``, e.g. ``("gzip", 2)``. Slows down writing a lot
        but the resulting files are potentially much smaller.
    """

    def __init__(
        self,
        file,
        mode="r",
        array_type="numpy",
        title=None,
        numDims=None,
        numNodes=None,
        numElems=None,
        numBlocks=None,
        numNodeSets=None,
        numSideSets=None,
        io_size=0,
        compression=None,
    ):

        if compression:
            self._comp_opts = {
                "chunks": True,
                "compression": compression[0],
                "compression_opts": compression[1],
            }
        else:
            self._comp_opts = {}

        # API is currently quite limited...mainly because nothing else is
        # implemented.
        assert mode in ["r", "a", "w"], "Only 'r', 'a', or 'w' is supported."
        assert array_type == "numpy", "array_type must be 'numpy'."

        if mode == "w":
            assert numNodeSets == 0, "numNodeSets must be 0 for now."
            assert numDims in [2, 3], "Only 2 or 3 dimensions are supported."

            # Determines the precision with which floating point variables are
            # written.
            if io_size == 0:
                if platform.architecture()[0] == "64bit":  # pragma: no cover
                    self.__f_dtype = np.float64
                    self.__f_word_size = 8
                else:  # pragma: no cover
                    self.__f_dtype = np.float32
                    self.__f_word_size = 4
            elif io_size == 4:
                self.__f_dtype = np.float32
                self.__f_word_size = 4
            elif io_size == 8:
                self.__f_dtype = np.float64
                self.__f_word_size = 8
            else:  # pragma: no cover
                raise NotImplementedError

            assert not os.path.exists(file), "File '%s' already exists." % file

            self._f = h5netcdf.File(file, mode="w")

            self._write_attrs(title=title)

            # Set the dimensions - very straightforward.
            # XXX: These should come from some header variable.
            self._f.dimensions["len_string"] = 33
            self._f.dimensions["len_line"] = 81
            # No clue what this is for...
            self._f.dimensions["four"] = 4
            self._f.dimensions["len_name"] = 256
            self._f.dimensions["time_step"] = None

            # These are dynamic.
            self._f.dimensions["num_dim"] = numDims
            self._f.dimensions["num_nodes"] = numNodes
            self._f.dimensions["num_elem"] = numElems
            self._f.dimensions["num_el_blk"] = numBlocks
            if numSideSets:
                self._f.dimensions["num_side_sets"] = numSideSets

            self._create_variables()

        elif mode in ["r", "a"]:
            if mode == "r":
                assert os.path.exists(file), "File '%s' does not exist." % file
            self._f = h5netcdf.File(file, mode=mode)

        else:  # pragma: no cover
            raise NotImplementedError

    @property
    def num_dims(self):
        """
        Number of dimensions in the exodus file.
        """
        return int(get_dim_size(self._f, "num_dim"))

    @property
    def num_el_blk(self):
        """
        Number of element blocks in the exodus file.
        """
        return int(get_dim_size(self._f, "num_el_blk"))

    def put_info_records(self, info):
        """
        Puts the info records into the exodus file.

        Does currently not do anything.

        :type info: list of str
        :param info: The strings to save.
        """
        if not info:
            return

        for _i, value in enumerate(info):
            assert len(value) < get_dim_size(
                self._f, "len_line"
            ), "Records '%s' is longer then %i letters." % (
                value,
                get_dim_size(self._f, "len_line"),
            )

        self._f.dimensions["num_info"] = len(info)

        self._f.create_variable(
            "info_records",
            ("num_info", "len_line"),
            dtype="|S1",
            **self._comp_opts
        )

        ir = self._f.variables["info_records"]
        for idx, value in enumerate(info):
            # Clear just to be safe.
            ir[idx] = b""
            if not value:
                continue
            ir[idx, : len(value)] = [
                _i.encode() if hasattr(_i, "encode") else _i for _i in value
            ]

    def put_coords(self, xCoords, yCoords, zCoords):
        """
        Put coordinates in the exodus file.

        :type xCoords: :class:`numpy.ndarray`
        :param xCoords: The X coordiantes.
        :type yCoords: :class:`numpy.ndarray`
        :param yCoords:  The Y coordinates.
        :type zCoords: :class:`numpy.ndarray`
        :param zCoords:  The Z coordinates.
        """
        self._f.variables["coordx"][:] = xCoords
        self._f.variables["coordy"][:] = yCoords
        self._f.variables["coordz"][:] = zCoords

    def put_elem_blk_info(
        self, id, elemType, numElems, numNodesPerElem, numAttrsPerElem
    ):
        """
        Set the details of an element block.

        :type id: int
        :param id: The id of the element block.
        :type elemType: str
        :param elemType: The name of the element block.
        :type numElems: int
        :param numElems: The number of elements in the block.
        :type numNodesPerElem: int
        :param numNodesPerElem: The number of nodes per element.
        :type numAttrsPerElem: int
        :param numAttrsPerElem: The number of attributes per element.
        """
        assert numElems <= get_dim_size(
            self._f, "num_elem"
        ), "Canont have more elements in the block then globally set."
        assert numAttrsPerElem == 0, "Must be 0 for now."

        # So the logic is as follows. `eb_status` keeps track of which
        # element ids have already been assigned. We find the first that is
        # not zero and that is the actual index of the the element block.
        status = self._f._variables["eb_status"][:]
        assert 0 in status, "All element blocks already set."
        idx = np.argwhere(status == 0)[0][0] + 1

        num_el_name = "num_el_in_blk%i" % idx
        num_node_per_el_name = "num_nod_per_el%i" % idx
        var_name = "connect%i" % idx

        self._f.dimensions[num_el_name] = numElems
        self._f.dimensions[num_node_per_el_name] = numNodesPerElem

        self._f.create_variable(
            var_name,
            (num_el_name, num_node_per_el_name),
            dtype=np.int32,
            **self._comp_opts
        )
        self._f.variables[var_name].attrs["elem_type"] = np.string_(elemType)

        # Set the status and thus "claim" the element block id.
        self._f.variables["eb_status"][idx - 1] = 1
        # For some reason this is always eb_prop1.
        self._f.variables["eb_prop1"][idx - 1] = id

    def put_elem_connectivity(
        self, id, connectivity, shift_indices=0, chunk_size_in_mb=128
    ):
        """
        Set the element connectivity array for all elements in a block.

        The information for this block must have already been set by calling
        e.put_elem_blk_info() on this block.

        :type id: int
        :param id: The id of the element block.
        :type connectivity: :class:`numpy.ndarray`
        :param connectivity: The connectivity. Must be equal to the number
            of of elements times the number of nodes per element for any
            given block.
        :type shift_indices: int
        :param shift_indices: **Not available in the official exodus Python
            API!** This value will be added to all indices before they are
            written to the file. This is useful if you for example
            internally work with a 0-based indexing scheme but exodus
            requires a 1-based indexing scheme.
        :type chunk_size_in_mb: int
        :param chunk_size_in_mb: If ``shift_indices`` != 0 values will be
            written in chunks of this size. This is also the maximum memory
            usage of this method. If ``shift_indices`` == 0, all indices
            will be written directly from memory and no additional memory is
            required.
        """
        num_el_name = "num_el_in_blk%i" % id
        num_node_per_el_name = "num_nod_per_el%i" % id
        var_name = "connect%i" % id

        assert num_el_name in self._f.dimensions, (
            "Block id %i does not exist" % id
        )
        assert num_node_per_el_name in self._f.dimensions, (
            "Block id %i does not exist" % id
        )

        assert connectivity.size == (
            get_dim_size(self._f, num_el_name)
            * get_dim_size(self._f, num_node_per_el_name)
        )

        if shift_indices:
            ne = get_dim_size(self._f, num_el_name)
            nn = get_dim_size(self._f, num_node_per_el_name)

            chunk_size = int(
                chunk_size_in_mb * 1024**2 / connectivity.dtype.itemsize / nn
            )

            _t = connectivity.reshape((ne, nn))

            idx = 0
            while idx < ne:
                self._f.variables[var_name][idx : idx + chunk_size] = (  # NOQA
                    _t[idx : idx + chunk_size] + shift_indices  # NOQA
                )
                idx += chunk_size
        else:
            self._f.variables[var_name][:] = connectivity.reshape(
                (
                    get_dim_size(self._f, num_el_name),
                    get_dim_size(self._f, num_node_per_el_name),
                )
            )

    def put_time(self, step, value):
        """
        Put time step and value into exodus file.

        :type step: int
        :param step: The index of the time step. First is 1.
        :type value: float
        :param value: The actual time at that index.
        """
        self.__resize_time_if_necessary(step)
        self._f.variables["time_whole"][step - 1] = value

    def set_global_variable_number(self, number):
        """
        Set number of global variables in exodus file.

        :type number: int
        :param number: The number of global variables.
        """
        if not number:  # pragma: no cover
            return

        self._f.dimensions["num_glo_var"] = number

        self._f.create_variable(
            "name_glo_var",
            ("num_glo_var", "len_name"),
            dtype="|S1",
            **self._comp_opts
        )
        self._f.create_variable(
            "vals_glo_var",
            ("time_step", "num_glo_var"),
            dtype=self.__f_dtype,
            **self._comp_opts
        )

    def put_global_variable_name(self, name, index):
        """
        Put global variable with name at index into exodus file.

        :type name: str
        :param name: The name of the variable.
        :type index: int
        :param index: The index of the global variable. First is 1!
        """
        self._f.variables["name_glo_var"][index - 1] = b""
        self._f.variables["name_glo_var"][index - 1, : len(name)] = [
            _i.encode() if hasattr(_i, "encode") else _i for _i in name
        ]

    def put_global_variable_value(self, name, step, value):
        """
        Put global variable value and variable name at time step into exodus
        file.

        :type name: str
        :param name: The name of the variable.
        :type step: int
        :param step: The index of the time step. First is 1.
        :type value: float
        :param value: The actual time at that index.
        """
        self.__resize_time_if_necessary(step)
        idx = self.get_global_variable_names().index(name)
        self._f.variables["vals_glo_var"][step - 1, idx] = value

    def get_global_variable_names(self):
        """
        Get list of global variable names in exodus file.
        """
        return [
            b"".join(_i).strip().decode()
            for _i in self._f.variables["name_glo_var"][:]
        ]

    def get_global_variable_values(self, name):
        """
        Get global variables values within an exodus file.
        """
        idx = self.get_global_variable_names().index(name)
        return self._f.variables["vals_glo_var"][0, idx]

    def set_element_variable_number(self, number):
        """
        Set number of element variables in exodus file.

        :type number: int
        :param number: The number of variables per element.
        """
        if not number:
            return

        self._f.dimensions["num_elem_var"] = number

        self._f.create_variable(
            "name_elem_var",
            ("num_elem_var", "len_name"),
            dtype="|S1",
            **self._comp_opts
        )

    def put_element_variable_name(self, name, index):
        """
        Element variable with name at index into exodus file.

        :type name: str
        :param name: The name of the element variable.
        :type index: int
        :param index: The index of the element variable. Starts with 1!
        """
        self._f.variables["name_elem_var"][index - 1] = b""
        self._f.variables["name_elem_var"][index - 1, : len(name)] = [
            _i.encode() if hasattr(_i, "encode") else _i for _i in name
        ]

    def get_element_variable_names(self):
        """
        Get list of element variable names in exodus file.
        """
        return [
            b"".join(_i).strip().decode()
            for _i in self._f.variables["name_elem_var"][:]
        ]

    def put_element_variable_values(self, blockId, name, step, values):
        """
        Put values into element block id and variable name at step.

        :type blockId: int
        :param blockId: The block id.
        :type name: str
        :param name: The name of the variable.
        :type step: int
        :param step: The time step at which to put the values.
        :type values: :class:`numpy.ndarray`
        :param values: The actual values.
        """
        self.__resize_time_if_necessary(step)

        num_elem_name = "num_el_in_blk%i" % blockId
        assert num_elem_name in self._f.dimensions, (
            "Block id %i not found." % blockId
        )

        # 1-based indexing!
        idx = self.get_element_variable_names().index(name) + 1

        variable_name = "vals_elem_var%ieb%i" % (idx, blockId)

        # If it does not exist, create it.
        if variable_name not in self._f.variables:
            self._f.create_variable(
                variable_name,
                ("time_step", num_elem_name),
                dtype=self.__f_dtype,
                **self._comp_opts
            )

        self._f.variables[variable_name][step - 1] = values

    def get_element_variable_values(self, blockId, name, step):
        """
        Get values from element block id and variable name at step.

        :type blockId: int
        :param blockId: The block id.
        :type name: str
        :param name: The name of the variable.
        :type step: int
        :param step: The time step at which to put the values.
        Return values: The actual values.
        """
        assert step > 0, "Step must be larger than 0."
        assert get_dim_size(
            self._f, "time_step"
        ) is None or step <= get_dim_size(self._f, "time_step")

        num_elem_name = "num_el_in_blk%i" % blockId
        assert num_elem_name in self._f.dimensions, (
            "Block id %i not found." % blockId
        )

        # 1-based indexing!
        idx = self.get_element_variable_names().index(name) + 1

        variable_name = "vals_elem_var%ieb%i" % (idx, blockId)

        # If it does not exist, raise exception
        assert variable_name in self._f.variables, (
            "Variable %s not found" % variable_name
        )

        return self._f.variables[variable_name][step - 1][:]

    def set_node_variable_number(self, number):
        """
        Set number of node variables in exodus file.

        :type number: int
        :param number: The number of node variables.
        """
        if number == 0:  # pragma: no cover
            return

        self._f.dimensions["num_nod_var"] = number

        self._f.create_variable(
            "name_nod_var",
            ("num_nod_var", "len_name"),
            dtype="|S1",
            **self._comp_opts
        )

        for _i in range(number):
            name = "vals_nod_var%i" % (_i + 1)
            self._f.create_variable(
                name,
                ("time_step", "num_nodes"),
                dtype=self.__f_dtype,
                **self._comp_opts
            )

    def put_node_variable_name(self, name, index):
        """
        Node variable with name at index into exodus file.

        :type name: str
        :param name: The name of the element variable.
        :type index: int
        :param index: The index of the element variable. Starts with 1!
        """
        # 1 - based indexing!
        assert index <= get_dim_size(self._f, "num_nod_var")

        self._f.variables["name_nod_var"][index - 1] = b""
        self._f.variables["name_nod_var"][index - 1, : len(name)] = [
            _i.encode() if hasattr(_i, "encode") else _i for _i in name
        ]

    def get_node_variable_names(self):
        """
        Get list of node variable names in exodus file.
        """
        return [
            b"".join(_i).strip().decode()
            for _i in self._f.variables["name_nod_var"][:]
        ]

    def get_node_variable_number(self):
        """
        Get number of node variables in the file.
        """
        if "num_nod_var" not in self._f.dimensions:
            return 0
        return int(get_dim_size(self._f, "num_nod_var"))

    def __resize_time_if_necessary(self, step):
        assert step > 0, "Step must be larger than 0."
        d = self._f.dimensions["time_step"]
        # Compatibility.
        if isinstance(d, (int, np.integer)) or d is None:
            assert d is None or step <= d
            if step > self._f._current_dim_sizes["time_step"]:
                self._f.resize_dimension("time_step", step)
        else:
            assert d is None or d.isunlimited()
            if step > d.size:
                self._f.resize_dimension("time_step", step)

    def put_node_variable_values(self, name, step, values):
        """
        Put node values into variable name at step into exodus file

        :type name: str
        :param name: The name of the variable.
        :type step: int
        :param step: The time step at which to put the values.
        :type values: :class:`numpy.ndarray`
        :param values: The actual values.
        """
        self.__resize_time_if_necessary(step)

        # 1-based indexing!
        idx = self.get_node_variable_names().index(name) + 1

        d_name = "vals_nod_var%i" % idx
        self._f.variables[d_name][step - 1] = values

    def get_node_variable_values(self, name, step):
        """
        Get the node variable values for a a certain step.

        :type name: str
        :param name: The name of the variable.
        :type step: int
        :param step: The time step at which to get the values.
        """
        # Make sure the step is valid.
        if step <= 0:
            raise ValueError("Step must be larger than zero.")
        # This cannot really happen anymore with newer pyexodus versions, as
        # the time_step is always an unlimited dimension so there is not easy
        # way to test this. But it might still happen if somebody reads a file
        # created with an older pyexodus version so I'll leave the block in
        # but mark it as uncovered by the tests.
        if get_dim_size(self._f, "time_step") is not None and not (
            0 < step <= get_dim_size(self._f, "time_step")
        ):  # pragma: no cover
            msg = "Step must be 0 < step < %i." % get_dim_size(
                self._f, "time_step"
            )
            raise ValueError(msg)
        # Will raise with a reasonable error message if name is not correct.
        # 1-based indexing!
        idx = self.get_node_variable_names().index(name) + 1

        d_name = "vals_nod_var%i" % idx
        # If it is resizeable, check the actual size.
        if get_dim_size(self._f, "time_step") is None:
            available_steps = self._f.variables[d_name].shape[0]
            if not (0 < step <= available_steps):
                msg = "Step must be 0 < step <= %i." % available_steps
                raise ValueError(msg)

        return self._f.variables[d_name][step - 1][:]

    def put_side_set_params(self, id, numSetSides, numSetDistFacts):
        """
        Set ID, num elements, and num nodes of a sideset

        :type id: int
        :param id: The id of the side set.
        :type numSetSides: int
        :param numSetSides: The number of elements for a side set.
        :type numSetDistFacts: int
        :param numSetDistFacts: The number of nodes for the side set.
        """
        assert numSetDistFacts == 0, "Only 0 is currently supported."

        assert id not in self._f.variables["ss_prop1"][:], (
            "Side set id %i already exists." % id
        )

        _t = self._f.variables["ss_status"][:]
        count = len(_t[_t > 0])
        assert count < get_dim_size(
            self._f, "num_side_sets"
        ), "Maximum number of side sets reached."

        idx = count + 1
        dim_name = "num_side_ss%i" % idx
        elem_ss_name = "elem_ss%i" % idx
        side_ss_name = "side_ss%i" % idx

        # Create the dimension and variables.
        self._f.dimensions[dim_name] = numSetSides
        self._f.create_variable(
            elem_ss_name, (dim_name,), dtype=np.int32, **self._comp_opts
        )
        self._f.create_variable(
            side_ss_name, (dim_name,), dtype=np.int32, **self._comp_opts
        )

        # Set meta-data.
        self._f.variables["ss_status"][idx - 1] = 1
        # For reasons I don't understand, this is ALWAYS ss_prop1.
        self._f.variables["ss_prop1"][idx - 1] = id

    def put_side_set(self, id, sideSetElements, sideSetSides):
        """
        Set the id, element ids, and side ids for a side set (creates the
        side set).

        :type id: int
        :param id: The id of the side set.
        :type sideSetElements: :class:`numpy.ndarray`
        :param sideSetElements: The side set elements.
        :type sideSetSides: :class:`numpy.ndarray`
        :param sideSetSides: The side set sides.
        """
        # Find the side set.
        _idx = self._f.variables["ss_prop1"][:]
        assert id in _idx, "Could not find side set with id %i." % id
        # 1-based indexing!
        idx = np.argwhere(_idx == id)[0][0] + 1

        elem_ss_name = "elem_ss%i" % idx
        side_ss_name = "side_ss%i" % idx

        self._f.variables[elem_ss_name][:] = sideSetElements
        self._f.variables[side_ss_name][:] = sideSetSides

    def put_side_set_name(self, id, name):
        """
        Write side set name for side set "id" in exodus file

        :type id: int
        :param id: The id of the side set.
        :type name: str
        :param name: The string of the side set.
        """
        # Find the side set.
        _idx = self._f.variables["ss_prop1"][:]
        assert id in _idx, "Could not find side set with id %i." % id
        # 1-based indexing!
        idx = np.argwhere(_idx == id)[0][0] + 1

        self._f.variables["ss_names"][idx - 1] = b""
        self._f.variables["ss_names"][idx - 1, : len(name)] = [
            _i.encode() if hasattr(_i, "encode") else _i for _i in name
        ]

    def get_side_set_names(self):
        """
        Get a list of the side set names in the exodus file.
        """
        _side_sets = self._f.variables["ss_names"][:]
        side_sets = []
        for _i in _side_sets:
            side_sets.append(
                "".join(
                    _j.decode() if hasattr(_j, "decode") else _j for _j in _i
                )
            )
        return side_sets

    def get_side_set_ids(self):
        """
        Get a list of side set ids in the exodus file.
        """
        return [int(_i) for _i in self._f.variables["ss_prop1"][:]]

    def get_side_set(self, id):
        """
        Get element and side ids for a certain side set.

        Returns a tuple of two arrays. The first are the element indices,
        the second the side ids in these elements.

        :type id: int
        :param id: The id of the side set.
        """
        ids = self.get_side_set_ids()
        if id not in ids:
            raise ValueError(
                "No side set with id %i in file. Available "
                "ids: %s." % (id, ", ".join(["%i" % _i for _i in ids]))
            )
        id = ids.index(id) + 1
        side_name = "side_ss%i" % id
        elem_name = "elem_ss%i" % id

        return self._f.variables[elem_name][:], self._f.variables[side_name][:]

    def get_elem_type_for_block(self, id):
        """
        Return the element type for an element block as a string.

        .. note::

            This method does not have a counter part in the official exodus
            Python API.

        :type id: int
        :param id: The element block id.
        """
        var_name = "connect%i" % id
        if var_name not in self._f.variables:
            raise ValueError("No element block with id %i in file." % id)
        elem_type = self._f[var_name].attrs["elem_type"]
        try:
            elem_type = elem_type.decode()
        # Not triggered on all Python versions.
        except AttributeError:  # pragma: no cover
            pass
        return elem_type

    def get_side_set_node_list(self, id):
        """
        Get the nodes for a certain side set.

        Returns a tuple of two arrays. The first contains the number of
        nodes on each face the the seconds are the local node ids for the
        side set in the exodus file.

        :type id: int
        :param id: The id of the side set.
        """
        # XXX: Currently only works for files with a single element block.
        elem_type = self.get_elem_type_for_block(id=1)

        elem_idx, side_idx = self.get_side_set(id=id)
        _sin = _SIDE_SET_NUMBERING[elem_type]

        num_nodes = np.ones_like(elem_idx) * _sin.shape[1]
        # This one is a bit tricky. Not sure if the current solution is
        # optimal but it gets the trick done and does avoid a bunch of copies.
        # Step 1: Get all elements in the side set. Account for 1-based
        # indexing
        _e = self._f.variables["connect1"][:][elem_idx - 1]
        # Step 2: From each element we have to pick these sides.
        _s = _sin[side_idx - 1]
        # This still has some additional allocations but otherwise indexes
        # pretty directly.
        local_node_ids = _e.ravel()[
            (_s + np.arange(_e.shape[0])[:, np.newaxis] * _e.shape[1]).ravel()
        ]

        return num_nodes, local_node_ids

    def get_coord(self, i):
        """
        Get x, y, z of i-th node in the exodus file.

        :type i: int or array/list of ints
        :param i: Node index or indices. 1-based. The official exodus API
            can only take single indices - pyexodus can take a list of indices.
            In that case it will still return a three-tuple, but each now
            contains multiple variables.
        """
        # Make it work with single indices and arrays.
        i = list(np.atleast_1d(i) - 1)
        if len(i) == 1:
            i = int(i[0])
            if not 1 <= i + 1 <= get_dim_size(self._f, "num_nodes"):
                raise ValueError(
                    "Invalid index. Coordinate bounds: [1, %i]."
                    % get_dim_size(self._f, "num_nodes")
                )

        x = self._f.variables["coordx"][i]
        y = self._f.variables["coordy"][i]
        if get_dim_size(self._f, "num_dim") == 2:
            return x, y, np.zeros_like(x)
        return x, y, self._f.variables["coordz"][i]

    def get_coords(self):
        """
        Returns all nodes in x, y, z.
        """
        x = self._f.variables["coordx"][:]
        y = self._f.variables["coordy"][:]
        if get_dim_size(self._f, "num_dim") == 2:
            return x, y, np.zeros_like(x)
        return x, y, self._f.variables["coordz"][:]

    def get_elem_connectivity(self, id, indices=None):
        """
        Get the connectivity for a certain element block.

        Returns a tuple with three things: The actual connectivity as an
        array of node indices, the number of elements in this block, and the
        number of nodes per elements in this block.

        :type id: int
        :param id: Id of the element block.
        :type indices: :class:`numpy.ndarray`
        :param indices: If given, only get the connectivity of the specified
            element indices relative to the element block. These are 1-based
            indices in accordance with the exodus convention! Note that the
            second and third returned item are always stats for the whole
            connectivity array regardless of this argument. This parameter
            is not part of the official exodus Python API.
        """
        var_name = "connect%i" % id
        conn = self._f.variables[var_name]

        # Read everything if indices is not given.
        if indices is None:
            indices = slice(None)
        else:
            indices = np.array(indices)
            indices = list(indices - 1)

        return conn[indices], conn.shape[0], conn.shape[1]

    def _write_attrs(self, title):
        """
        Write all the attributes.
        """
        self._f.attrs["api_version"] = np.float32([7.05])
        self._f.attrs["version"] = np.float32([7.05])
        self._f.attrs["floating_point_word_size"] = np.array(
            [self.__f_word_size], dtype=np.int32
        )
        self._f.attrs["file_size"] = np.array([1], dtype=np.int32)
        self._f.attrs["maximum_name_length"] = np.array([32], dtype=np.int32)
        self._f.attrs["int64_status"] = np.array([0], dtype=np.int32)
        self._f.attrs["title"] = np.string_(title)

    def _create_variables(self):
        # Time steps.
        self._f.create_variable(
            "/time_whole",
            ("time_step",),
            dtype=self.__f_dtype,
            **self._comp_opts
        )

        # Element block stuff.
        self._f.create_variable(
            "/eb_names",
            ("num_el_blk", "len_name"),
            dtype="|S1",
            **self._comp_opts
        )
        self._f.create_variable(
            "/eb_status", ("num_el_blk",), dtype=np.int32, **self._comp_opts
        )
        # I don't really understand the number here yet...
        self._f.create_variable(
            "/eb_prop1",
            ("num_el_blk",),
            dtype=np.int32,
            data=[-1] * get_dim_size(self._f, "num_el_blk"),
            **self._comp_opts
        )
        self._f.variables["eb_prop1"].attrs["name"] = np.string_("ID")

        # Coordinate names.
        self._f.create_variable(
            "/coor_names",
            ("num_dim", "len_name"),
            dtype="|S1",
            **self._comp_opts
        )

        # Coordinates.
        for i in "xyz":
            self._f.create_variable(
                "/coord" + i,
                ("num_nodes",),
                dtype=self.__f_dtype,
                **self._comp_opts
            )

        # Side sets.
        if "num_side_sets" in self._f.dimensions:
            self._f.create_variable(
                "/ss_names",
                ("num_side_sets", "len_name"),
                dtype="|S1",
                **self._comp_opts
            )
            self._f.create_variable(
                "/ss_prop1",
                ("num_side_sets",),
                dtype=np.int32,
                data=[-1] * get_dim_size(self._f, "num_side_sets"),
                **self._comp_opts
            )
            self._f.variables["ss_prop1"].attrs["name"] = np.string_("ID")
            self._f.create_variable(
                "/ss_status",
                ("num_side_sets",),
                dtype=np.int32,
                **self._comp_opts
            )

    def __del__(self):
        try:
            self._f.close()
        except Exception:  # pragma: no cover
            pass

    def close(self):
        try:
            self._f.close()
        except Exception:  # pragma: no cover
            pass

    def __enter__(self):
        """
        Enable usage as a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Enable usage as a context manager.
        """
        self.__del__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    Lion Krischer (lionkrischer@gmail.com), 2016
:license:
    MIT License
"""
from __future__ import absolute_import

from .core import exodus  # NOQA
from .version import get_git_version

__version__ = get_git_version()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Setup script for the pyexodus module.

:copyright:
    Lion Krischer (lion.krischer@mondaic.com), 2016-2024
:license:
    MIT License
"""
import inspect
import os
import sys

from setuptools import setup, find_packages


# Import the version string.
path = os.path.join(
    os.path.abspath(os.path.dirname(inspect.getfile(inspect.currentframe()))),
    "pyexodus",
)
sys.path.insert(0, path)
from version import get_git_version  # NOQA


def get_package_data():
    """
    Returns a list of all files needed for the installation relative to the
    'pyexodus' subfolder.
    """
    return ["RELEASE-VERSION"]


setup_config = dict(
    name="pyexodus",
    version=get_git_version(),
    description="Module for creating Exodus files",
    long_description=(
        "Create Exodus files with Python "
        "(https://salvushub.github.io/pyexodus/)"
    ),
    author="Lion Krischer",
    author_email="lionkrischer@gmail.com",
    url="https://github.com/SalvusHub/pyexodus",
    packages=find_packages(),
    license="MIT",
    platforms="OS Independent",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    install_requires=["numpy", "h5netcdf >= 0.5.0"],
    package_data={"pyexodus": get_package_data()},
)


if __name__ == "__main__":
    setup(**setup_config)



# -*- coding: utf-8 -*-
# Author: Douglas Creager <dcreager@dcreager.net>
# This file is placed into the public domain.
#
# Modified by the ObsPy project and pyexodus.

# Calculates the current version number.  If possible, this is the
# output of “git describe”, modified to conform to the versioning
# scheme that setuptools uses.  If “git describe” returns an error
# (most likely because we're in an unpacked copy of a release tarball,
# rather than in a git working copy), then we fall back on reading the
# contents of the RELEASE-VERSION file.
#
# To use this script, simply import it your setup.py file, and use the
# results of get_git_version() as your package version:
#
# from version import *
#
# setup(
#     version=get_git_version(),
#     .
#     .
#     .
# )
#
# This will automatically update the RELEASE-VERSION file, if
# necessary.  Note that the RELEASE-VERSION file should *not* be
# checked into git; please add it to your top-level .gitignore file.
#
# You'll probably want to distribute the RELEASE-VERSION file in your
# sdist tarballs; to do this, just create a MANIFEST.in file that
# contains the following line:
#
#   include RELEASE-VERSION
# NO IMPORTS FROM PYEXODUS OR FUTURE! (file gets used at
# installation time)


import io
import os
import inspect
from subprocess import Popen, PIPE

__all__ = "get_git_version"


script_dir = os.path.abspath(
    os.path.dirname(inspect.getfile(inspect.currentframe()))
)
PYEXODUS_ROOT = os.path.abspath(os.path.join(script_dir, os.pardir))
VERSION_FILE = os.path.join(script_dir, "RELEASE-VERSION")


def call_git_describe(abbrev=4):  # pragma: no cover
    try:
        p = Popen(
            ["git", "rev-parse", "--show-toplevel"],
            cwd=PYEXODUS_ROOT,
            stdout=PIPE,
            stderr=PIPE,
        )
        p.stderr.close()
        path = p.stdout.readline().decode().strip()
        p.stdout.close()
    except Exception:
        return None
    if os.path.normpath(path) != PYEXODUS_ROOT:
        return None
    try:
        p = Popen(
            [
                "git",
                "describe",
                "--dirty",
                "--abbrev=%d" % abbrev,
                "--always",
                "--tags",
            ],
            cwd=PYEXODUS_ROOT,
            stdout=PIPE,
            stderr=PIPE,
        )

        p.stderr.close()
        line = p.stdout.readline().decode()
        p.stdout.close()

        if "-" not in line and "." not in line:
            line = "0.0.0-g%s" % line
        return line.strip()
    except Exception:
        return None


def read_release_version():  # pragma: no cover
    try:
        with io.open(VERSION_FILE, "rt") as fh:
            version = fh.readline()
        return version.strip()
    except Exception:
        return None


def write_release_version(version):  # pragma: no cover
    with io.open(VERSION_FILE, "wb") as fh:
        fh.write(("%s\n" % version).encode("ascii", "strict"))


def get_git_version(abbrev=4):  # pragma: no cover
    # Read in the version that's currently in RELEASE-VERSION.
    release_version = read_release_version()

    # First try to get the current version using “git describe”.
    version = call_git_describe(abbrev)

    # If that doesn't work, fall back on the value that's in
    # RELEASE-VERSION.
    if version is None:
        version = release_version

    # If we still don't have anything, that's an error.
    if version is None:
        return "0.0.0-tar/zipball"

    # If the current version is different from what's in the
    # RELEASE-VERSION file, update the file to be current.
    if version != release_version:
        write_release_version(version)

    # Finally, return the current version.
    return version
